unit modMain;

{' /*******************************************************************************
'   modMain.pas within DelphiSpec.dpr
'
'   Public variable declarations, startup and initialization code,
'   and routines for loading the ROM and snapshots
'
'   Original author (vbSpec): Chris Cowley <ccowley@grok.co.uk>
'   Copyright (C)1999-2000 Grok Developments Ltd.
'   http://www.grok.co.uk/
'
'   Translation to Delphi Object Pascal by
'   Jari Korhonen <jarit.korhonen@luukku.com>
'   Copyright (C) 2001- Jari Korhonen
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/}


interface

uses Windows, Graphics, Mmsystem, IniFiles, Sysutils, Forms, Extctrls, Math, DIB;

// Flag bit positions
Const
  F_C = 1;
  F_N = 2;
  F_PV = 4;
  F_3 = 8;
  F_H = 16;
  F_5 = 32;
  F_Z = 64;
  F_S = 128;
  BORDERWIDTH = 20;
  DELPHISPEC : String = 'DelphiSpec';
  Z80RUNNING=1;       //Normally
  Z80PAUSED=2;        //When in setting screens
  Z80STOPPED=3;       //When losing focus

// Screen things
  nPixelsWide = 256;
  nPixelsHigh = 192;
  nCharsWide = 32;
  nCharsHigh = 24;
  FirstAttr = nPixelsHigh * nCharsWide;
  //lastAttr = FirstAttr + (nCharsHigh * nCharsWide);


var
glTstatesPerInterrupt: integer;
Parity: array[0..256] of Boolean;
glInterruptTimer: integer;

// Memory handling variables for 128K spectrum emulation
glMemPagingType:integer;    // 0 = No mem paging (48K speccy)
                            // 1 = 128K/+2 paging available
                            // 2 = 128K/+2 and +2A/+3 special paging available

gRAMPage: array[0..11, 0..16383] of Byte; //(pages 0 to 7 are RAM, and pages 8 to 11 are ROM)
glPageAt: array[0..4] of integer;
glLastOut7FFD: integer;
glLastOut1FFD: integer;
glUseScreen: integer;
glEmulatedModel: integer;

{$IFDEF USE_SOUND}
gbSoundEnabled: Boolean;
glSoundRegister: integer;       // Contains currently indexed AY-3-8912 sound register
glBufNum: integer;              // ID of the last Wave buffer used
{$ENDIF}

// Main Z80 registers
regA: integer;
regHL: integer;
regB: integer;
regC: integer;
regDE: integer;

// Z80 Flags
fS: Boolean;
fZ: Boolean;
f5: Boolean;
fH: Boolean;
f3: Boolean;
fPV: Boolean;
fN: Boolean;
fC: Boolean;


// Alternate registers
regAF_: integer;
regHL_: integer;
regBC_: integer;
regDE_: integer;

// Index registers  - ID used as temp for ix/iy
regIX: integer;
regIY: integer;
regID: integer;

// Stack pointer and program counter
regSP: integer;
regPC: integer;

// Interrupt registers and flip-flops, and refresh registers
intI: integer;
intR: integer;
intRTemp: integer;
intIFF1: Boolean;
intIFF2: Boolean;
intIM: integer;

// Addressable Memory (16KB ROM at 0x0000, 48KB RAM at 0x4000)
// We also place a shadow copy of the ROM above the addressable memory space
// at 0x10000 so we don't have to check for regPC > 65535 at each instruction
mem: array[0..81919] of Byte;


ScrnLines: array[0..nPixelsHigh - 1, 0..32] of Boolean;
ScrnNeedRepaint: Boolean;
bFlashInverse: Boolean;
glScreenMem: array[0..191, 0..31] of integer;


// Interrupts/Screen refreshing
interruptCounter: integer;
glInterruptDelay: integer;
glDelayOverage: integer;

// Keypresses
keyB_SPC: integer;
keyH_ENT: integer;
keyY_P: integer;
key6_0: integer;
key1_5: integer;
keyQ_T: integer;
keyA_G: integer;
keyCAPS_V: integer;


glDisplayHeight: integer;
glDisplayWidth: integer;
glDisplayMultiplier: integer;


{$IFDEF USE_SOUND}
glBeeperVal: integer;
{$ENDIF}

function InitializeWaveOut : Boolean;
function GetSetting(settingName: string; defaultValue:string): string;
procedure SaveSetting(settingName: string; value:string);

procedure ForceRepaint;
procedure SetDisplaySize(lWidth : integer; lHeight : integer);
procedure SetEmulatedModel(lModel : integer);
procedure SetStatus(sMsg : String);
procedure Main;
function LoadROM(sROMFile : String = 'spectrum.rom'; lROMPage : integer = 8): boolean;
procedure LoadZ80Snap(sFilename : String);
procedure LoadSNASnap(sFilename : String);
procedure CloseWaveOut;
procedure resetKeyboard;
procedure emptyRAMs;
function ramCrc(page: integer): integer;
function GetAppPath: string;
procedure SaveROM(sFilename : string);
procedure SaveSNASnap(sFilename : String);
procedure SaveZ80Snap(sFilename : String);

implementation

uses Spectrum, modSpectrum, modAY8912, modWaveOut, modZ80;


function ramCrc(page: integer): integer;
  var Crc, ii : integer;
begin
  Crc := 0;
  for ii := 0 to 16383 do
    Crc := Crc + gRAMPage[page, ii];
  ramCRC := Crc;
end;

procedure emptyRAMs;
  var ii, jj: integer;
begin
  for ii := 0 to 11 do
    for jj := 0 to 16383 do
      gRAMPage[ii, jj] := 0;
end;



//Function emulates VB's App.Path
function GetAppPath: string;
begin
  GetAppPath := ExtractFilePath(Application.Exename);
end;


//Function emulates VB's GetSetting API with TIniFile
function GetSetting(settingName: string; defaultValue:string): string;
  var IniFile: TIniFile;
begin
  IniFile := TIniFile.Create('DELPHISPEC.INI');
  GetSetting := IniFile.ReadString('DelphiSpec', settingName, defaultvalue);
  IniFile.Free;
end;

//Function emulates VB's SaveSetting API with TIniFile
procedure SaveSetting(settingName: string; value:string);
  var IniFile: TIniFile;
begin
  IniFile := TIniFile.Create('DELPHISPEC.INI');
  IniFile.WriteString('DelphiSpec', settingName, value);
  IniFile.Free;
end;


procedure CloseWaveOut;
{$IFDEF USE_SOUND}
  var lRet : integer;
{$ENDIF}
begin
    {$IFDEF USE_SOUND}
        waveOutReset(glphWaveOut);
        For lRet := 1 To NUM_WAV_BUFFERS do
            waveOutUnprepareHeader(glphWaveOut,
              @gtWavHdr[lRet], sizeof(WAVEHDR));

        For lRet := 1 To NUM_WAV_BUFFERS do
        begin
            GlobalUnlock(ghMem[lRet]);
            GlobalFree(ghMem[lRet]);
        end;

        waveOutClose(glphWaveOut);

        gbSoundEnabled := False;
    {$ENDIF}
End;

Function CompressMemoryBlock(lBlock : integer; pData : pByte) : integer;
    var pDataStart: pByte; ii: integer; lCounter : integer; lCounter2 : integer; lByteCount : integer;
begin
    pDataStart := pData;
    lCounter := 0;
    repeat
        If gRAMPage[lBlock, lCounter] = gRAMPage[lBlock, lCounter + 1] Then
        begin
          // String of two or more bytes starting from lCounter
          // Might be a candidate for compression
          lCounter2 := lCounter;
          While (lCounter2 - lCounter) < 260 do
          begin
              if (lCounter2 >= 16383) then
                break;
                
              If gRAMPage[lBlock, lCounter2 + 1] = gRAMPage[lBlock, lCounter2] Then
                  lCounter2 := lCounter2 + 1
              Else
                  break;
          end;
          lByteCount := lCounter2 - lCounter + 1;
          lCounter := lCounter2 + 1;
          If lByteCount > 255 Then
          begin
              lCounter := lCounter - (lByteCount - 255);
              lByteCount := 255;
          End;

          If lByteCount > 4 Then
          begin
              pData^ := $ED;
              inc(pData);
              pData^ := $ED;
              inc(pData);
              pData^ := lByteCount;
              inc(pData);
              pData^ := gRAMPage[lBlock, lCounter - 1];
              inc(pData);
              If lCounter < 16384 Then
              begin
                  pData^ := gRAMPage[lBlock, lCounter];
                  inc(pData);
                  lCounter := lCounter + 1;
              End;
          end
          Else
          begin
              If gRAMPage[lBlock, lCounter - 1] = $ED Then
              begin
                  // Strings of 2 or more ED bytes are always encoded
                  pData^ := $ED;
                  inc(pData);
                  pData^ := $ED;
                  inc(pData);
                  pData^ := lByteCount;
                  inc(pData);
                  pData^ := gRAMPage[lBlock, lCounter - 1];
                  inc(pData);
                  If lCounter < 16384 Then
                  begin
                    pData^ := gRAMPage[lBlock, lCounter];
                    inc(pData);
                    lCounter := lCounter + 1;
                  End;
              end
              Else
              begin
                  // Strings of 4 or fewer bytes are not encoded
                  for ii := 1 to lByteCount do
                  begin
                    pData^ := gRAMPage[lBlock, lCounter - 1];
                    inc(pData);
                  end;
              end;
          End;
        end
        Else
        begin
          // A single byte, just write it out
          pData^ := gRAMPage[lBlock, lCounter];
          inc(pData);
          lCounter := lCounter + 1;
        End;
    Until lCounter >= 16383;

    If lCounter = 16383 Then
    begin;
        // Write out the final byte
        pData^ := gRAMPage[lBlock, 16383];
        inc(pData);
    end;

    lCounter2 := Integer(pData);
    lCounter := Integer(pDataStart);
    CompressMemoryBlock := lCounter2-lCounter;
End;

Function GetFilePart(sFilename : String) : String;
begin
    GetFilePart := ExtractFileName(sFileName);
End;


Function InitializeWaveOut : Boolean;
{$IFDEF USE_SOUND}
  var lRet : integer; sErrMsg : String; lCounter : integer; lCounter2 : integer;
{$ENDIF}
begin
    {$IFDEF USE_SOUND}
        If StrToInt(GetSetting('SoundEnabled', '1')) = 0 Then
        begin
            InitializeWaveOut := False;
            Exit;
        End;

        With gtWavFormat do
        begin
            wFormatTag := WAVE_FORMAT_PCM;
            nChannels := 1;
            nSamplesPerSec := WAVE_FREQUENCY;
            nAvgBytesPerSec := WAVE_FREQUENCY;
            nBlockAlign := 1;
            wBitsPerSample := 8;
            cbSize := 0;
        End;

        lRet := waveOutOpen(@glphWaveOut, WAVE_MAPPER, @gtWavFormat, 0, 0, CALLBACK_NULL);
        If lRet <> MMSYSERR_NOERROR Then
        begin
            SetLength(sErrMsg,  256);
            waveOutGetErrorText(lRet, PChar(sErrMsg), 255);
            Application.MessageBox(
              PChar('Error initialising WaveOut device.' + #13#10#13#10 + sErrMsg),
              PChar(DELPHISPEC),
              MB_OK or MB_ICONEXCLAMATION);
            InitializeWaveOut := False;
            Exit;
        End;

        For lCounter := 0 To NUM_WAV_BUFFERS do
        begin
            ghMem[lCounter] := GlobalAlloc(GPTR, WAV_BUFFER_SIZE+100);
            gpMem[lCounter] := GlobalLock(ghMem[lCounter]);
            gtWavHdr[lCounter].lpData := gpMem[lCounter];
            gtWavHdr[lCounter].dwBufferLength := WAV_BUFFER_SIZE;
            gtWavHdr[lCounter].dwUser := 0;
            gtWavHdr[lCounter].dwFlags := 0;
            gtWavHdr[lCounter].dwLoops := 0;
            gtWavHdr[lCounter].lpNext := nil;

            lRet := waveOutPrepareHeader(glphWaveOut,
              @gtWavHdr[lCounter],
              sizeof(WAVEHDR));

            If lRet <> MMSYSERR_NOERROR Then
            begin
                SetLength(sErrMsg,  256);
                waveOutGetErrorText(lRet, PChar(sErrMsg), 255);
                Application.MessageBox(
                  PChar('Error preparing wave header.' + #13#10#13#10 + sErrMsg),
                  PChar(DELPHISPEC),
                  MB_OK or MB_ICONEXCLAMATION);
                waveOutClose(glphWaveOut);
                For lCounter2 := 0 To NUM_WAV_BUFFERS do
                begin
                    GlobalUnlock(ghMem[lCounter2]);
                    GlobalFree(ghMem[lCounter2]);
                end;
                InitializeWaveOut := False;
                Exit;
            End;
        end;
        InitializeWaveOut := True;
     {$ELSE}
        InitializeWaveOut := True;
    {$ENDIF}
End;


procedure initParity;
    var lCounter : integer; j : Byte; p : Boolean;
    var powval: integer;
begin
    For lCounter := 0 To 255 do
    begin
        p := True;
        For j := 0 To 7 do
        begin
            PowVal := Trunc(IntPower(2, j));
            If (lCounter And powVal) <> 0 Then p := Not p;
        end;
        Parity[lCounter] := p;
    end;
End;

//Force repaint of whole spectrum screen
procedure ForceRepaint;
    var i : integer; X : integer;
begin
    ScrnNeedRepaint := True;
    For i := 0 To 191 do
    begin
        ScrnLines[i, 32] := True;
        For X := 0 To 31 do
        begin
            ScrnLines[i, X] := True;
        end;
    end;
End;


procedure InitScreenMemTable;
    var X : integer; Y : integer;
begin
    For Y := 0 To 191 do
    begin
        For X := 0 To 31 do
        begin
            glScreenMem[Y, X] := ((((Y div 8) * 32) + (Y Mod 8) * 256) + ((Y div 64) * 2048) - (Y div 64) * 256) + X;
        end;
    end;
End;


//Return value True=Success, False=Failure
function LoadROM(sROMFile : String = 'spectrum.rom'; lROMPage : integer = 8): boolean;
  var
    hFile : File;
    ret: integer;
    sROM : array[0..65536] of Byte;
    lCounter : integer;
    sr: TSearchRec;

  begin
    LoadROM := True;
    If FindFirst(sROMFile, 0, sr) <> 0 Then
    begin
      Application.MessageBox(
        PChar('The ROM image file "' + sROMFile + '" could not be found.'),
        PChar(DELPHISPEC),
        MB_OK or MB_ICONEXCLAMATION);

      LoadROM := false;
      Exit;
    End;

    AssignFile(hFile, sRomFile);
    reset(hFile, 1);

    // Read the ROM image into sROM
    Blockread(hFile, sROM, 16384, ret);
    If ret <> 16384 Then
    begin
      Application.MessageBox(
        PChar('An error ocurred whilst reading the ROM image file: "' + sROMFile + '"'),
        PChar(DELPHISPEC),
        MB_OK or MB_ICONEXCLAMATION);

      CloseFile(hFile);

      LoadRom := false;
      Exit;
    End
    else
    begin
{$IFDEF USE_DEBUGGING}
      ReportDebugInfo('LoadROM Start, Page ' + IntToStr(lROMPage), 0, 7, True);
{$ENDIF}
      // Copy the ROM into the appropriate memory page
      For lCounter := 1 To 16384 do
      begin
        gRAMPage[lROMPage, lCounter - 1] := sROM[lCounter-1];
      end;
      CloseFile(hFile);
{$IFDEF USE_DEBUGGING}
      ReportDebugInfo('LoadROM End, Page ' + IntToStr(lROMPage), 0, 7, True);
{$ENDIF}
    end;
end;


procedure LoadSNA128Snap(var hFile : File);
  var
    sData : array[0..65535] of byte;
    sTemp : array[0..1] of byte;
    lOut7FFD : integer;
    lBank : integer;
    lCounter : integer;
begin
    // Read first three banks
    BlockRead(hFile, sData, 49152);

    // PC
    BlockRead(hFile, sTemp, 2);
    regPC := 256*sTemp[1]+sTemp[0];

    // Last out to 0x7FFD
    BlockRead(hFile, sTemp, 1);
    lOut7FFD := sTemp[0];

    // Is TR-DOS paged? (ignored by vbSpec)
    BlockRead(hFile, sTemp, 1);

    //Setup first three banks
    For lCounter := 0 To 16383 do
    begin
        gRAMPage[5, lCounter] := sData[lCounter];
        gRAMPage[2, lCounter] := sData[lCounter + 16384];
        gRAMPage[lOut7FFD And 7, lCounter] := sData[lCounter + 32768];
    end;

    lBank := 0;
    While lBank < 8 do
    begin
        BlockRead(hFile, sData, 16384);
        If (lBank = 5) Or (lBank = 2) Or (lBank = (lOut7FFD And 7)) Then
            lBank := lBank + 1;

        If lBank < 8 Then
        begin
            For lCounter := 0 To 16383 do
                gRAMPage[lBank, lCounter] := sData[lCounter];

            lBank := lBank + 1;
        End;
    end;

    outb($7FFD, lOut7FFD);
End;

procedure resetKeyboard;
begin
    keyB_SPC := $FF;
    keyH_ENT := $FF;
    keyY_P := $FF;
    key6_0 := $FF;
    key1_5 := $FF;
    keyQ_T := $FF;
    keyA_G := $FF;
    keyCAPS_V := $FF;
End;

procedure LoadSNASnap(sFilename : String);
    var hFile : File; sData : array[0..65535] of byte; iCounter : integer; temp:integer;

begin
    AssignFile(hFile, sFileName);
    reset(hFile, 1);
    BlockRead(hFile, sData, 1);
    intI := sData[0];

    BlockRead(hFile, sData, 2);
    regHL_ := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regDE_ := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regBC_ := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regAF_ := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regHL := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regDE := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    setBC(256*sData[1]+sData[0]);

    BlockRead(hFile, sData, 2);
    regIY := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    regIX := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 1);
    If (sData[0] And 4) = 4 Then
    begin
      intIFF1 := True;
      intIFF2 := True;
    end
    Else
    begin
      intIFF1 := False;
      intIFF2 := False;
    End;

    BlockRead(hFile, sData, 1);
    intR := sData[0];
    intRTemp := intR;

    BlockRead(hFile, sData, 2);
    setAF(256*sData[1]+sData[0]);

    BlockRead(hFile, sData, 2);
    regSP := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 1);
    intIM := sData[0];

    BlockRead(hFile, sData, 1);
    // Border color
    glNewBorder := sData[0] And $7;

    Temp := FileSize(hFile);
    If Temp > 49180 Then
    begin
        LoadSNA128Snap(hFile);
        CloseFile(hFile);

        ForceRepaint;

        // Set the initial border color
        frmMainWnd.borderColor := glNewBorder;

        SetEmulatedModel(1);

        frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);

        screenPaint;
        resetKeyboard;
    end
    Else
    begin
        // Load a 48K Snapshot file
        BlockRead(hFile, sData, 49152);
        CloseFile(hFile);

{$IFDEF USE_DEBUGGING}
        ReportDebug('LoadSNA_0', 0, True);
{$ENDIF}
        For iCounter := 0 To 16383 do
        begin
            gRAMPage[5, iCounter] := sData[iCounter];
            gRAMPage[1, iCounter] := sData[iCounter + 16384];
            gRAMPage[2, iCounter] := sData[iCounter + 32768];
        end;

{$IFDEF USE_DEBUGGING}
        ReportDebug('LoadSNA_1', 0, True);
{$ENDIF}
        ForceRepaint;

        // Set the initial border color
        frmMainWnd.borderColor := glNewBorder;

        SetEmulatedModel(0);
{$IFDEF USE_DEBUGGING}
        ReportDebug('LoadSNA_2', 0, True);
{$ENDIF}

        frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);

        screenPaint;
        resetKeyboard;
        poppc;
    End;
End;



procedure ReadZ80V2orV3Snap(var hFile : File);
  var
  lHeaderLen : integer; sData : array[0..65535] of byte;
  lCounter : integer; bHardwareSupported : Boolean;
  lMemPage : integer; lBlockCounter : integer; b128K : Boolean; lMemPos : integer;
  lOutFFFD : integer; sModel: string; size:integer;
begin
    lCounter := 0;
    BlockRead(hFile, sData, 2);
    lHeaderLen := 256 * sData[1] + sData[0];
    b128k := false;

    // offset 32 - 2 bytes - PC
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 2);
        regPC := 256 * sData[1] + sdata[0];
        lCounter := lCounter + 2;
    End;

    // offset 34 - 1 byte - hardware mode
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        Case sData[0] of
          0:    // 48K spectrum
            begin
              bHardwareSupported := True;
              SetEmulatedModel(0);
            end;
          1:  //48K spectrum + Interface 1
            begin
              bHardwareSupported := True;
              SetEmulatedModel(0);
            end;
          2:  //SamROM
            begin
              sModel := 'SamROM';
              bHardwareSupported := False;
            end;
          3:
            begin
              If lHeaderLen = 23 Then
              begin
                sModel := '128K Spectrum';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              end
              Else
              begin
                sModel := '48K Spectrum + M.G.T.';
                bHardwareSupported := True;
                SetEmulatedModel(0);
              End;
            end;
          4:
            begin
              If lHeaderLen = 23 Then
              begin
                sModel := '128K Spectrum + Interface 1';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              end
              Else
              begin
                sModel := '128K Spectrum';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              End;
            end;
          5:
            begin
              bHardwareSupported := True;
              sModel := '128K Spectrum + Interface 1';
              b128K := True;
              SetEmulatedModel(1);
            end;
          6:
            begin
              bHardwareSupported := False;
              sModel := '128K Spectrum + M.G.T.';
              b128K := True;
              SetEmulatedModel(1);
            end;
          7:
            begin
              bHardwareSupported := False;
              sModel := 'ZX Spectrum +3';
              b128K := True;
              SetEmulatedModel(2);
            end;
          Else
          begin
            bHardwareSupported := False;
            sModel := 'Unknown hardware platform';
          end;
        End;

        If bHardwareSupported = False Then
        begin
            Application.MessageBox(
              PChar('DelphiSpec does not support the required hardware platform (' + sModel + ') for this snapshot. This snapshot may not function correctly in DelphiSpec.'),
              PChar(DELPHISPEC),
              MB_OK);
        End;

        lCounter := lCounter + 1;
    End;

    // offset 35 - 1 byte - last out to 0x7FFD - not required for 48K spectrum
    If lCounter < lHeaderLen Then
    begin
        If b128K Then
        begin
          BlockRead(hFile, sData, 1);
          outb($7FFD, sData[0]);
        end
        Else
        begin
            BlockRead(hFile, sData, 1);
            glPageAt[0] := 8;
            glPageAt[1] := 5;
            glPageAt[2] := 1;
            glPageAt[3] := 2;
            glPageAt[4] := 8;
        End;
        lCounter := lCounter + 1;
    End;

    // offset 36 - 1 byte - 0xFF if Interface 1 ROM is paged in
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lCounter := lCounter + 1;

        If sData[0] = 255 Then
            Application.MessageBox(
              PChar('This snapshot was saved with the Interface 1 ROM paged in. It may not run correctly in DelphiSpec.'),
              PChar(DELPHISPEC),
              MB_OK);
    End;

    // offset 37 - 1 byte (bit 0: 1=intR emulation on, bit 1: 1=LDIR emulation on
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lCounter := lCounter + 1;
    End;

    // offset 38 - Last out to 0xFFFD (+2/+3 sound chip register number)
    lOutFFFD := 0;
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lOutFFFD := sData[0];
        lCounter := lCounter + 1;
    end;

    // offset 39 - 16 bytes - contents of the sound chip registers
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 16);
        If b128K Then
        begin
            AYWriteReg(0, sData[0]);
            AYWriteReg(1, sData[1]);
            AYWriteReg(2, sData[2]);
            AYWriteReg(3, sData[3]);
            AYWriteReg(4, sData[4]);
            AYWriteReg(5, sData[5]);
            AYWriteReg(6, sData[6]);
            AYWriteReg(7, sData[7]);
            AYWriteReg(8, sData[8]);
            AYWriteReg(9, sData[9]);
            AYWriteReg(10, sData[10]);
            AYWriteReg(11, sData[11]);
            AYWriteReg(12, sData[12]);
            AYWriteReg(13, sData[13]);
            AYWriteReg(14, sData[14]);
            AYWriteReg(15, sData[15]);
        end;
        lCounter := lCounter + 16;
    End;

    If b128K Then
        outb($FFFD, lOutFFFD);

    // read the remaining bytes of the header (we don't care what information they hold)
    If lCounter < lHeaderLen Then
        BlockRead(hFile, sData, lHeaderLen-lCounter);

    repeat
        // read a block
        BlockRead(hFile, sData, 2);
        If EOF(hFile) Then break;
        lHeaderLen := 256 * sData[1] + sData[0];
        BlockRead(hFile, sData, 1);
        Case sData[0] of
          0:  // Spectrum ROM
            If b128K Then lMemPage := 9 Else lMemPage := 8;
          1:  //Interface 1 ROM, or similar (we discard these blocks)
            lMemPage := -1;
          2:  //128K ROM (reset)
            If b128K Then lMemPage := 8 Else lMemPage := -1;
          3:  //Page 0 (not used by 48K Spectrum)
            If b128K Then lMemPage := 0 Else lMemPage := -1;
          4:  //Page 1 RAM at 0x8000
            lMemPage := 1;
          5:  //Page 2 RAM at 0xC000
            lMemPage := 2;
          6:  //Page 3 (not used by 48K Spectrum)
            If b128K Then lMemPage := 3 Else lMemPage := -1;
          7:  //Page 4 (not used by 48K Spectrum)
            If b128K Then lMemPage := 4 Else lMemPage := -1;
          8:  //Page 5 RAM at 0x4000
            lMemPage := 5;
          9:  //Page 6 (not used by 48K Spectrum)
            If b128K Then lMemPage := 6 Else lMemPage := -1;
          10: //Page 7 (not used by 48K Spectrum)
            If b128K Then lMemPage := 7 Else lMemPage := -1;
          11: //Multiface ROM
            lMemPage:=-1;
          else
            lMemPage:=-1;
        End;

        BlockRead(hFile, sData, lHeaderLen, size);

        If lMemPage <> -1 Then
        begin
            If lHeaderLen = $FFFF Then
            begin
                //Not a compressed block, just copy it straight into RAM
                For lCounter := 0 To 16383 do
                    gRAMPage[lMemPage, lCounter] := sData[lCounter];
            end
            Else
            begin
                // Uncompress the block to memory
                lCounter := 0;
                lMemPos := 0;
                repeat
                    If sData[lCounter] = $ED Then
                    begin
                        If sData[lCounter + 1] = $ED Then
                        begin
                            // This is an encoded block
                            lCounter := lCounter + 2;
                            lHeaderLen := sData[lCounter];
                            lCounter := lCounter + 1;
                            For lBlockCounter := 0 To lHeaderLen - 1 do
                                gRAMPage[lMemPage, lMemPos + lBlockCounter] := sData[lCounter];

                            lMemPos := lMemPos + lHeaderlen;
                        end
                        Else
                        begin
                            // Just a single ED, write it out
                            gRAMPage[lMemPage, lMemPos] := $ED;
                            lMemPos := lMemPos + 1;
                        End;
                    end
                    Else
                    begin
                        gRAMPage[lMemPage, lMemPos] := sData[lCounter];
                        inc(lMemPos);
                    End;
                    inc(lCounter);
                until lCounter >= size;
            End;
        End;
    Until Eof(hFile);
End;


procedure ReadZ80V1Snap(var hFile : File);
    var
    lDataLen : integer; sData : array[0..65535] of Byte; lBlockLen : integer;
    lCounter : integer; lMemPos : integer; lBlockCounter : integer;
    size : integer;
begin
    lDataLen := FileSize(hFile)-FilePos(hFile);

    //read the compressed data into sData
    BlockRead(hFile, sData, lDataLen, size);

    // Uncompress the block to memory
    lCounter := 0;
    lMemPos := 16384;
    repeat
        If sData[lCounter] = $ED Then
        begin
            If sData[lCounter + 1] = $ED Then
            begin
                // This is an encoded block
                lCounter := lCounter + 2;
                lBlockLen := sData[lCounter];
                inc(lCounter);
                For lBlockCounter := 1 To lBlockLen do
                begin
                    gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := sData[lCounter];
                    inc(lMemPos);
                end
            end
            Else
            begin
                // Just a single ED, write it out
                gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := $ED;
                inc(lMemPos);
            End
        end
        Else
        begin
            gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := sData[lCounter];
            inc(lMemPos);
        End;
        inc(lCounter);
    Until lCounter > size - 4;

    If (sData[lCounter-1] <> 0)
      or (sData[lCounter] <> $ED)
      or (sData[lCounter+1] <> $ED)
      or (sData[lCounter+2] <> 0) Then
        Application.MessageBox(
              PChar('Error in compressed Z80 file. Block end marker 0x00EDED00 is not present.'),
              PChar(DELPHISPEC),
              MB_OK or MB_ICONEXCLAMATION);

End;




procedure LoadZ80Snap(sFilename : String);
    var
      hFile: File; sData : array[0..65535] of byte; iCounter : integer;
      bCompressed : Boolean;
begin
    AssignFile(hFile, sFileName);
    reset(hFile, 1);

{$IFDEF USE_DEBUGGING}
    ReportDebug('LoadZ80_0',0,True);
{$ENDIF}
    glPageAt[0] := 8;
    glPageAt[1] := 5;
    glPageAt[2] := 2;
    glPageAt[3] := 0;
    glPageAt[4] := 8;

    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then AY8912_reset;
    {$ENDIF}

    //byte 0 - A register
    BlockRead(hFile, sData, 1);
    regA := sData[0];

    //byte 1 - F register
    BlockRead(hFile, sData, 1);
    setF(sData[0]);

    //bytes 2 + 3 - BC register pair (C first, then B)
    BlockRead(hFile, sData, 2);
    setBC(256*sData[1]+sData[0]);

    //bytes 4 + 5 - HL register pair
    BlockRead(hFile, sData, 2);
    regHL := 256*sData[1]+sData[0];

    //bytes 6 + 7 - PC (this is zero for v2.x or v3.0 Z80 files)
    BlockRead(hFile, sData, 2);
    regPC := 256*sData[1]+sData[0];

    //bytes 8 + 9 - SP
    BlockRead(hFile, sData, 2);
    regSP := 256*sData[1]+sData[0];

    //byte 10 - Interrupt register
    BlockRead(hFile, sData, 1);
    intI := sData[0];

    //byte 11 - Refresh register
    BlockRead(hFile, sData, 1);
    intR := (sData[0] and 127);

    //byte 12 - bitfield
    BlockRead(hFile, sData, 1);
    //if byte 12 = 255 then it must be treated as if it = 1, for compatibility with other emulators
    If sData[0] = 255 Then sData[0] := 1;
    //bit 0 - bit 7 of R
    If (sData[0] And 1) = 1 Then intR := intR Or 128;
    intRTemp := intR;

    //bits 1,2 and 3 - border color
    glNewBorder := (sData[0] And 14) div 2;
    //bit 4 - 1 if SamROM switched in (we don't care about this!)
    //bit 5 - if 1 and PC<>0 then the snapshot is compressed using the
    //         rudimentary Z80 run-length encoding scheme
    bCompressed:=false;
    If (sData[0] And $20) = $20 Then bCompressed := True;
    //bits 6 + 7 - no meaning

    //bytes 13 + 14 - DE register pair
    BlockRead(hFile, sData, 2);
    regDE := 256*sData[1]+sData[0];

    //bytes 15 + 16 - BC' register pair
    BlockRead(hFile, sData, 2);
    regBC_ := 256*sData[1]+sData[0];

    //bytes 17 + 18 - DE' register pair
    BlockRead(hFile, sData, 2);
    regDE_ := 256*sData[1]+sData[0];

    //bytes 19 + 20 - HL' register pair
    BlockRead(hFile, sData, 2);
    regHL_ := 256*sData[1]+sData[0];

    //bytes 21 + 22 - AF' register pair (A first then F - not Z80 byte order!!)
    BlockRead(hFile, sData, 2);
    regAF_ := 256*sData[0]+sData[1];

    //byte 23 + 24 - IY register pair
    BlockRead(hFile, sData, 2);
    regIY := 256*sData[1]+sData[0];

    //byte 25 + 26 - IX register pair
    BlockRead(hFile, sData, 2);
    regIX := 256*sData[1]+sData[0];

    //byte 27 - Interrupt flipflop (0=DI, else EI)
    BlockRead(hFile, sData, 1);
    If sData[0] = 0 Then
    begin
        intIFF1 := False;
        intIFF2 := False;
    end
    Else
    begin
        intIFF1 := True;
        intIFF2 := True;
    End;
    //byte 28 - IFF2 (ignored)
    BlockRead(hFile, sData, 1);

    //byte 29 - Interrupt mode (bits 2 - 7 contain info about joystick modes etc, which we ignore)
    BlockRead(hFile, sData, 1);
    intIM := sData[0] And 3;
{$IFDEF USE_DEBUGGING}
    ReportDebug('LoadZ80_1',0,True);
{$ENDIF}

    If regPC = 0 Then
    begin
        //This is a V2 or V3 Z80 file
        ReadZ80V2orV3Snap(hFile);
        CloseFile(hFile);
    end
    Else
    begin
        SetEmulatedModel(0); // V1 .Z80 snapshots are all 48K
        // PC<>0, so lets check to see if this is a compressed V1 Z80 file
        If bCompressed Then
        begin
            //Uncompress the RAM data
            ReadZ80V1Snap(hFile);
            CloseFile(hFile);
        end
        Else
        begin
            // Uncompressed Z80 file
            BlockRead(hFile, sData, 49152);
            CloseFile(hFile);

            // Copy the RAM data to addressable memory space
            For iCounter := 16384 To 65535 do
              gRAMPage[glPageAt[iCounter div 16384], iCounter And 16383] := sData[iCounter - 16384];
        End;
    End;
{$IFDEF USE_DEBUGGING}
    ReportDebug('LoadZ80_2',0,True);
{$ENDIF}

    ForceRepaint;

    // Set the initial border color
    frmMainWnd.borderColor := (glNewBorder and 7);

    frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);

    ForceRepaint;
    screenPaint;
    resetKeyboard;
End;


function SpectrumColorTable: TRGBQuads;
var
  i: Integer;
begin
  for i:=0 to 255 do
  begin
    with Result[i] do
    begin
      case i of
        0:
          begin
            rgbRed := 0;
            rgbGreen := 0;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        1:
          begin
            rgbRed := 0;
            rgbGreen := 0;
            rgbBlue := 192;
            rgbReserved := 0;
          end;
        2:
          begin
            rgbRed := 192;
            rgbGreen := 0;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        3:
          begin
            rgbRed := 192;
            rgbGreen := 0;
            rgbBlue := 192;
            rgbReserved := 0;
          end;
        4:
          begin
            rgbRed := 0;
            rgbGreen := 192;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        5:
          begin
            rgbRed := 0;
            rgbGreen := 192;
            rgbBlue := 192;
            rgbReserved := 0;
          end;
        6:
          begin
            rgbRed := 192;
            rgbGreen := 192;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        7:
          begin
            rgbRed := 192;
            rgbGreen := 192;
            rgbBlue := 192;
            rgbReserved := 0;
          end;
        8:
          begin
            rgbRed := 0;
            rgbGreen := 0;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        9:
          begin
            rgbRed := 0;
            rgbGreen := 0;
            rgbBlue := 255;
            rgbReserved := 0;
          end;
        10:
          begin
            rgbRed := 255;
            rgbGreen := 0;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        11:
          begin
            rgbRed := 255;
            rgbGreen := 0;
            rgbBlue := 255;
            rgbReserved := 0;
          end;
        12:
          begin
            rgbRed := 0;
            rgbGreen := 255;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        13:
          begin
            rgbRed := 0;
            rgbGreen := 255;
            rgbBlue := 255;
            rgbReserved := 0;
          end;
        14:
          begin
            rgbRed := 255;
            rgbGreen := 255;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
        15:
          begin
            rgbRed := 255;
            rgbGreen := 255;
            rgbBlue := 255;
            rgbReserved := 0;
          end;
        else
          begin
            rgbRed := 0;
            rgbGreen := 0;
            rgbBlue := 0;
            rgbReserved := 0;
          end;
      end;
    end;
  end;
end;



procedure Main;
  var ii: integer;
begin
    emptyRAMs;
    InitScreenMemTable;

    frmMainWnd.DXPaintBoxSpecScreen.DIB.ColorTable :=SpectrumColorTable;

    glDisplayWidth := StrToInt(GetSetting('DisplayWidth', '256'));
    glDisplayHeight := StrToInt(GetSetting('DisplayHeight', '192'));
    SetDisplaySize(glDisplayWidth, glDisplayHeight);

{$IFDEF USE_SOUND}
        AY8912_init(1773000, WAVE_FREQUENCY, 8);
{$ENDIF}

    glInterruptDelay := StrToInt(GetSetting('InterruptDelay', '20'));

    // Initialize everything
    initParity;
    Z80Reset;

    SetEmulatedModel(StrToInt(GetSetting('EmulatedModel', '0')));

    ForceRepaint;
    resetKeyboard;

    timeBeginPeriod(1);

    glInterruptTimer := timeGetTime() + 20;
    {$IFDEF USE_SOUND}
        gbSoundEnabled := InitializeWaveOut();
        If gbSoundEnabled Then
        begin
            For ii := 1 To 12 do
              waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(WAVEHDR));

            glBufNum := 12;
        End;
    {$ENDIF}
End;

Procedure SaveROM(sFilename : String);
  var hFile : File; lCounter : integer;
begin
  Assign(hFile, sFileName);
  rewrite(hFile, 1);
  for lCounter := 0 to 16383 do
    BlockWrite(hFile, gRAMPage[glPageAt[0], lCounter], 1);

  CloseFile(hFile);
End;

procedure SaveSNA128Snap(var hFile : File);
  var lBank : integer; lCounter : integer;
  var bDummy: byte; wDummy: word;
begin
  BlockWrite(hFile, intI, 1);
  BlockWrite(hFile, regHL_, 2);
  BlockWrite(hFile, regDE_, 2);
  BlockWrite(hFile, regBC_, 2);
  BlockWrite(hFile, regAF_, 2);

  BlockWrite(hFile, regHL, 2);
  BlockWrite(hFile, regDE, 2);
  BlockWrite(hFile, regC, 1);
  BlockWrite(hFile, regB, 1);
  BlockWrite(hFile, regIY, 2);
  BlockWrite(hFile, regIX, 2);

  //Interrupt flipflops
  If intIFF1 = True Then
    bdummy := 4
  Else
    bdummy := 0;
  BlockWrite(hFile, bdummy, 1);

  //R
  intRTemp := intRTemp And 127;
  bdummy := (intR And $80) Or intRTemp;
  BlockWrite(hFile, bdummy, 1);

  // AF
  wDummy := getAF;
  BlockWrite(hFile, wdummy, 2);

  // SP
  BlockWrite(hFile, regSP, 2);

  // Interrupt Mode
  BlockWrite(hFile, intIM, 1);

  bDummy := frmMainWnd.BorderColor;
  BlockWrite(hFile, bDummy, 1);

  // Save the three currently-paged RAM banks
  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[1], lCounter], 1);

  For lCounter := 0 To 16383 do
    BlockWrite(hFile, gRAMPage[glPageAt[2], lCounter], 1);

  For lCounter := 0 To 16383 do
    BlockWrite(hFile, gRAMPage[glPageAt[3], lCounter], 1);

  //PC
  BlockWrite(hFile, regPC, 2);

  //Last out to 0x7FFD
  BlockWrite(hFile, glLastOut7FFD, 1);

  //Is TR-DOS paged? (0=not paged, 1=paged)
  bDummy := 0;
  BlockWrite(hFile, bDummy, 1);

  // Save the remaining RAM banks
  lBank := 0;
  While lBank < 8 do
  begin
    If (lBank <> glPageAt[1]) And (lBank <> glPageAt[2]) And (lBank <> glPageAt[3]) Then
    begin
      For lCounter := 0 To 16383 do
        BlockWrite(hFile, gRAMPage[lBank, lCounter], 1);

      lBank := lBank + 1;
    end;
  end;
End;

procedure SaveSNASnap(sFilename : String);
    var hFile : File; lCounter : integer; bDummy: byte; wDummy: word;
begin

  Assign(hFile, sFileName);
  rewrite(hFile, 1);

  If glEmulatedModel <> 0 Then
  begin
      // We're not running in 48K mode
      SaveSNA128Snap(hFile);
      frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);
      CloseFile(hFile);
      Exit;
  end;
  pushpc;

  BlockWrite(hFile, intI, 1);
  BlockWrite(hFile, regHL_, 2);
  BlockWrite(hFile, regDE_, 2);
  BlockWrite(hFile, regBC_, 2);
  BlockWrite(hFile, regAF_, 2);

  BlockWrite(hFile, regHL, 2);
  BlockWrite(hFile, regDE, 2);
  BlockWrite(hFile, regC, 1);
  BlockWrite(hFile, regB, 1);
  BlockWrite(hFile, regIY, 2);
  BlockWrite(hFile, regIX, 2);

  //Interrupt flipflops
  If intIFF1 = True Then
    bDummy := 4
  Else
    bDummy := 0;
  BlockWrite(hFile, bDummy, 1);

  //R
  intRTemp := intRTemp And 127;
  bDummy := (intR And $80) Or intRTemp;
  BlockWrite(hFile, bDummy, 1);

  //AF
  wDummy := getAF;
  BlockWrite(hFile, wDummy, 2);

  //SP
  BlockWrite(hFile, regSP, 2);

  //Interrupt Mode
  BlockWrite(hFile, intIM, 1);

  BlockWrite(hFile, frmMainWnd.borderColor, 1);

  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[1], lCounter], 1);
  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[2], lCounter], 1);
  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[3], lCounter], 1);

  frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);
  CloseFile(hFile);
  poppc;
End;

procedure SaveZ80Snap(sFilename : String);
    var hFile : File; sData : array[0..65535] of byte;
      lCounter : integer; lBufSize : integer; bDummy:byte;
begin
  Assign(hFile, sFileName);
  rewrite(hFile, 1);

  //Index0:A,F
  BlockWrite(hFile, regA, 1);
  bDummy := getF;
  BlockWrite(hFile, bDummy, 1);

  //Index2:BC
  BlockWrite(hFile, regC, 1);
  BlockWrite(hFile, regB, 1);

  //Index4:HL
  BlockWrite(hFile, regHL, 2);

  //Index6
  //Set PC to zero to indicate a v3.0 Z80 file
  bDummy := 0;
  BlockWrite(hFile, bDummy, 1);
  BlockWrite(hFile, bDummy, 1);

  //Index8
  //SP
  BlockWrite(hFile, regSP, 2);

  //Index10
  //I
  BlockWrite(hFile, intI, 1);

  //Index11
  //R (7 bits)
  intRTemp := intRTemp And $7F;
  BlockWrite(hFile, intRTemp, 1);

  //Index12
  //bitfield
  if intR and $80 = $80 then
    bDummy := 1
  else
    bDummy := 0;
  bDummy := bDummy or (frmMainWnd.borderColor * 2);
  BlockWrite(hFile, bDummy, 1);

  //Index13
  //DE
  BlockWrite(hFile, regDE, 2);

  //Index15
  //BC'
  BlockWrite(hFile, regBC_, 2);

  //Index17
  //DE'
  BlockWrite(hFile, regDE_, 2);

  //Index19
  //HL'
  BlockWrite(hFile, regHL_, 2);

  //Index21
  //AF'
  bDummy := regAF_ div 256;
  BlockWrite(hFile, bDummy, 1);
  bDummy := regAF_ and 255;
  BlockWrite(hFile, bDummy, 1);

  //Index23
  //IY
  BlockWrite(hFile, regIY, 2);

  //Index25
  //IX
  BlockWrite(hFile, regIX, 2);

  //Index27
  //Interrupt flipflops
  If intIFF1 = True Then
  begin
    bDummy := 255;
    bDummy := 255;
  end
  Else
  begin
    bDummy := 0;
    bDummy := 0;
  End;
  BlockWrite(hFile, bDummy, 1);
  BlockWrite(hFile, bDummy, 1);

  //Index29
  // Interrupt Mode
  BlockWrite(hFile, intIM, 1);

  //Index30
  // V2.01 info
  bDummy := 23; BlockWrite(hFile, bDummy, 1);
  bDummy := 0; BlockWrite(hFile, bDummy, 1);

  //Index32
  //PC
  BlockWrite(hFile, regPC, 2);

  //Index34
  //Hardware mode
  If glEmulatedModel = 0 Then
  begin
      bDummy := 0;    //48K Spectrum
      BlockWrite(hFile, bDummy, 1);
      BlockWrite(hFile, bDummy, 1);
  end
  Else
  begin
      bDummy := 3;    //128K Spectrum
      BlockWrite(hFile, bDummy, 1);

      //Last out to 7FFD
      BlockWrite(hFile, glLastOut7FFD, 1);
  End;

  //Index36
  //IF.1 Paged in
  bDummy := 0; BlockWrite(hFile, bDummy, 1);

  //Index37
  //1=R emulation on, 2=LDIR emulation on
  bDummy := 3; BlockWrite(hFile, bDummy, 1);

  //Index38
  {$IFDEF USE_SOUND}
  bDummy := glSoundRegister; BlockWrite(hFile, bDummy, 1);
  {$ELSE}
  bDummy := 0; BlockWrite(hFile, bDummy, 1);
  {$ENDIF}

  //Index39
  //AY-3-8912 register contents
  For lCounter := 0 To 15 do
    BlockWrite(hFile, AYPSG.Regs[lCounter], 1);

  //Index54
  If glEmulatedModel = 0 Then
  begin
    // Block 1
    lBufSize := CompressMemoryBlock(glPageAt[1], Addr(sData));


    // Buffer length
    BlockWrite(hFile, lBufSize, 2);


    //Index49
    //Block number
    bDummy := glPageAt[1] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);

    //Block 2
    lBufSize := CompressMemoryBlock(glPageAt[2], Addr(sData));
    //Buffer length
    BlockWrite(hFile, lBufSize, 2);
    //Block number
    bDummy := glPageAt[2] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);

    //Block 3
    lBufSize := CompressMemoryBlock(glPageAt[3], Addr(sData));
    //Buffer length
    BlockWrite(hFile, lBufSize, 2);
    //Block number
    bDummy := glPageAt[3] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);
  end
  Else
  begin
      For lCounter := 0 To 7 do
      begin
        lBufSize := CompressMemoryBlock(lCounter, Addr(sData));
        //Buffer length
        BlockWrite(hFile, lBufSize, 2);
        //Block number
        bDummy := lCounter + 3;
        BlockWrite(hFile, bDummy, 1);
        BlockWrite(hFile, sData, lBufSize);
      end;
  End;

  frmMainWnd.Caption := DELPHISPEC + ' - ' + GetFilePart(sFilename);
  CloseFile(hFile);
End;

procedure SetDisplaySize(lWidth : integer; lHeight : integer);
  var Y: integer;
begin
  Y := frmMainWnd.statusBar.Height;

  glDisplayWidth := lWidth;
  glDisplayHeight := lHeight;
  SaveSetting('DisplayWidth', IntToStr(glDisplayWidth));
  SaveSetting('DisplayHeight', IntToStr(glDisplayHeight));

  frmMainWnd.DXPaintBoxSpecScreen.DIB.SetSize(glDisplayWidth, glDisplayHeight, 8);

  Case lWidth of
    256:  // Standard
      begin
        frmMainWnd.ClientWidth := 256;
        frmMainWnd.ClientHeight := 192+Y;
        glDisplayMultiplier := 1
      end;
    512:
      begin
        frmMainWnd.ClientWidth := 512;
        frmMainWnd.ClientHeight := 384+Y;
        glDisplayMultiplier := 2;
      end;
    768:
      begin
        frmMainWnd.ClientWidth := 768;
        frmMainWnd.ClientHeight := 576+Y;
        glDisplayMultiplier := 3;
      end;
  End;
  ForceRepaint;
End;


procedure SetEmulatedModel(lModel : integer);
    var sModel : String;
begin
    glEmulatedModel := lModel;

    Case lModel of
      0:
        begin
          // A 48K Spectrum has 69888 tstates per interrupt (3.50000 MHz)
          sModel := 'ZX Spectrum 48K';
          glTstatesPerInterrupt := 69888;

          {$IFDEF USE_SOUND}
          glWaveAddTStates := 158;
          {$ENDIF}
          glMemPagingType := 0;
          glUseScreen := 5;

          glPageAt[0] := 8;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := 8;
          // load the ROM image into memory
          if LoadROM(GetAppPath + 'spectrum.rom', 8) = false then
            Application.Terminate;
        end;
      1:
        begin
          // A 128K Spectrum has 70908 tstates per interrupt (3.54690 MHz)
          sModel := 'ZX Spectrum 128';
          glTstatesPerInterrupt := 70908;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}

          glMemPagingType := 1;
          glUseScreen := 5;

          glPageAt[0] := 8;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0;
          glPageAt[4] := 8;
          if LoadROM(GetAppPath + 'zx128_1.rom', 9) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'zx128_0.rom', 8) = false then
            Application.Terminate;
        end;
      2:
        begin
          // A Spectrum +2 has 70908 tstates per interrupt (3.54690 MHz)
          sModel := 'ZX Spectrum +2';
          glTstatesPerInterrupt := 70908;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}
          glMemPagingType := 1;
          glUseScreen := 5;

          glPageAt[0] := 8;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0;
          glPageAt[4] := 8;
          if LoadROM(GetAppPath + 'plus2_1.rom', 9) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus2_0.rom', 8) = false then
            Application.Terminate;
        end;
      3:  // +2A
        begin
          sModel := 'ZX Spectrum +2A';
          glTstatesPerInterrupt := 70908;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}

          glMemPagingType := 2;
          glUseScreen := 5;

          glPageAt[0] := 8;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := 8;

          if LoadROM(GetAppPath + 'plus2a_3.rom', 11) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus2a_2.rom', 10) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus2a_1.rom', 9) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus2a_0.rom', 8) = false then
            Application.Terminate;
        end;
      4:  // +3
        begin
          sModel := 'ZX Spectrum +3';
          glTstatesPerInterrupt := 70908;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}
          glMemPagingType := 2;
          glUseScreen := 5;

          glPageAt[0] := 8;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := 8;

          if LoadROM(GetAppPath + 'plus3_3.rom', 11) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus3_2.rom', 10) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus3_1.rom', 9) = false then
            Application.Terminate;
          if LoadROM(GetAppPath + 'plus3_0.rom', 8) = false then
            Application.Terminate;
        end;
    end;

    SetStatus(sModel);
End;


procedure SetStatus(sMsg : String);
begin
    frmMainWnd.statusBar.simpleText := sMsg;
End;

end.






